/**
 * EasyBeans
 * Copyright (C) 2007 Bull S.A.S.
 * Contact: easybeans@objectweb.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * --------------------------------------------------------------------------
 * $Id: DeployerFactory.java 1330 2007-03-13 10:49:59Z benoitf $
 * --------------------------------------------------------------------------
 */

package org.objectweb.easybeans.deployable;

import org.objectweb.easybeans.api.EZBServer;
import org.objectweb.easybeans.deployer.api.EZBDeployer;
import org.objectweb.easybeans.deployer.api.EZBDeployerException;
import org.objectweb.easybeans.log.JLog;
import org.objectweb.easybeans.log.JLogFactory;

/**
 * Allow to get a deployer. When EasyBeans is integrated into JOnAS, it will
 * return a JOnASDeployer, for EasyBeans in standalone mode an
 * EasyBeansDeployer, etc.
 * @author Florent Benoit
 */
public final class DeployerFactory {

    /**
     * Name of the property thats defines the deployer's class. (optional)
     */
    public static final String DEPLOYER_FACTORY_CLASS_NAME = DeployerFactory.class.getName();

    /**
     * Default deployer.
     */
    private static final String DEFAULT_DEPLOYER = "org.objectweb.easybeans.deployer.EasyBeansDeployer";

    /**
     * Logger.
     */
    private static JLog logger = JLogFactory.getLog(DeployerFactory.class);

    /**
     * Name of the class to use for the deployer.
     */
    private static String className = null;

    /**
     * Utility class, no public constructor.
     */
    private DeployerFactory() {

    }


    /**
     * Gets a deployer depending of the property or the class that has been
     * registered.
     * @param embedded the instance of the server.
     * @return an instance of the deployer.
     * @throws EZBDeployerException if the deployer cannot be returned.
     */
    @SuppressWarnings("unchecked")
    public static EZBDeployer getDeployer(final EZBServer embedded) throws EZBDeployerException {
        EZBDeployer deployer = null;

        // If class name is empty, set to the default value
        if (className == null) {
            className = DEFAULT_DEPLOYER;
        }

        // Use the system property of the class name if the sys property is not
        // set.
        String clName = System.getProperty(DEPLOYER_FACTORY_CLASS_NAME, className);

        logger.debug("Using ''{0}'' as deployer class", clName);

        // load the class
        Class<?> clazz;
        try {
            clazz = Thread.currentThread().getContextClassLoader().loadClass(clName);
        } catch (ClassNotFoundException e) {
            throw new EZBDeployerException("Cannot load the deployer class named '" + clName + "'.", e);
        }

        // Cast
        Class<EZBDeployer> depClass = null;
        if (EZBDeployer.class.isAssignableFrom(clazz)) {
            depClass = (Class<EZBDeployer>) clazz;
        } else {
            throw new EZBDeployerException("The class '" + clName + "' is not an EZBDeployer class");
        }


        // Build new instance
        try {
            deployer = depClass.newInstance();
        } catch (InstantiationException e) {
            throw new EZBDeployerException("Cannot instantiate the class '" + clName + "'.");
        } catch (IllegalAccessException e) {
            throw new EZBDeployerException("Cannot instantiate the class '" + clName + "'.");
        }

        // Set the embedded
        deployer.setEmbedded(embedded);

        // return
        return deployer;
    }


    /**
     * Set the name of the deployer class to use.
     * @param classNameStr the name of the class to use.
     */
    public static void setClassName(final String classNameStr) {
        DeployerFactory.className = classNameStr;
    }

}
