/**
 * EasyBeans
 * Copyright (C) 2007 Bull S.A.S.
 * Contact: easybeans@objectweb.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * --------------------------------------------------------------------------
 * $Id: DirectoryDeployerMonitor.java 1408 2007-05-15 17:03:49Z benoitf $
 * --------------------------------------------------------------------------
 */

package org.objectweb.easybeans.server;

import static org.objectweb.easybeans.util.url.URLUtils.urlToFile;

import java.io.File;
import java.net.URL;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.WeakHashMap;
import java.util.concurrent.ConcurrentHashMap;

import org.objectweb.easybeans.api.EZBArchive;
import org.objectweb.easybeans.api.EZBArchiveException;
import org.objectweb.easybeans.api.EZBContainer;
import org.objectweb.easybeans.api.EZBContainerException;
import org.objectweb.easybeans.deployable.DeployableHelper;
import org.objectweb.easybeans.deployable.DeployableHelperException;
import org.objectweb.easybeans.deployable.DeployerFactory;
import org.objectweb.easybeans.deployable.archive.ArchiveManager;
import org.objectweb.easybeans.deployer.api.EJB3Deployable;
import org.objectweb.easybeans.deployer.api.EZBDeployable;
import org.objectweb.easybeans.deployer.api.EZBDeployer;
import org.objectweb.easybeans.deployer.api.EZBDeployerException;
import org.objectweb.easybeans.log.JLog;
import org.objectweb.easybeans.log.JLogFactory;

/**
 * This class monitors scan the directory and deploy archive if any. Also,
 * archives are monitored. If there is a change, the archive is reloaded or
 * undeployed and then deployed.
 * @author Florent Benoit
 */
public class DirectoryDeployerMonitor extends ContainersMonitor {

    /**
     * Sleep time for the thread of the cleaner (5s).
     */
    private static final int SLEEP_TIME = 5000;

    /**
     * Logger.
     */
    private JLog logger = JLogFactory.getLog(DirectoryDeployerMonitor.class);

    /**
     * Map between an File (monitored) and the last updated file.
     */
    private Map<File, Long> modifiedFiles = null;

    /**
     * List of deployed files (by this monitor).
     */
    private Map<File, EZBDeployable> deployed = null;

    /**
     * List of File that have a failed deployment (by this monitor).
     */
    private List<File> failed = null;

    /**
     * Initializing period ?.
     */
    private boolean bootInProgress = false;

    /**
     * Stop order received ?
     */
    private boolean stopped = false;

    /**
     * Deployer.
     */
    private EZBDeployer deployer;

    /**
     * Builds a new monitor by initializing lists.
     * @param embedded the embedded server which is monitored.
     * @throws EmbeddedException if there is an exception for this monitor.
     */
    public DirectoryDeployerMonitor(final Embedded embedded) throws EmbeddedException {
        super(embedded);
        this.modifiedFiles = new WeakHashMap<File, Long>();
        this.deployed = new ConcurrentHashMap<File, EZBDeployable>();
        this.failed = new ArrayList<File>();


        // Set the name of this thread
        setName(this.getClass().getSimpleName());

        // Get the deployer.
        try {
            deployer = DeployerFactory.getDeployer(embedded);
        } catch (EZBDeployerException e) {
            throw new EmbeddedException("Cannot get a deployer", e);
        }

    }

    /**
     * Init containers (call at startup).
     */
    @Override
    public void init() {
        // Start the boot process.
        bootInProgress = true;

        // Initialize containers
        try {
            detectNewArchives();
        } catch (EmbeddedException e) {
            logger.error("Cannot scan for new archives", e);
        }

        // No more in the boot process.
        bootInProgress = false;
    }

    /**
     * Start the thread of this class It will clean all the work entries.
     */
    @Override
    public void run() {

        for (;;) {
            // Stop the thread
            if (stopped) {
                return;
            }

            // Check if existing EJB3 container have not been modified ?
            for (EZBContainer container : getEmbedded().getContainers().values()) {
                if (container.isAvailable()) {
                    checkContainer(container);
                }
            }

            // Check new archives/containers to start
            try {
                detectNewArchives();
            } catch (Exception e) { // Catch all exception (including runtime)
                logger.error("Problem when trying to find and deploy new archives", e);
            }

            // Undeploy/ReDeploy archives for deployed modules
            try {
                checkModifiedDeployables();
            } catch (Exception e) { // Catch all exception (including runtime)
                logger.error("Problem when checking current deployables", e);
            }

            // Thread wait
            try {
                Thread.sleep(SLEEP_TIME);
            } catch (InterruptedException e) {
                throw new RuntimeException("Thread fail to sleep");
            }
        }
    }

    /**
     * Scan all files present in the deploy directory and deploy them. (if not
     * deployed).
     * @throws EmbeddedException if there is a problem during the scan
     */
    private void detectNewArchives() throws EmbeddedException {

        // Get the list of deploy directories
        List<File> deployDirectories = getEmbedded().getServerConfig().getDeployDirectories();

        for (File deployDirectory : deployDirectories) {
            // get files
            File[] files = deployDirectory.listFiles();

            // next directory if there are no files to scan.
            if (files == null) {
                continue;
            }

            // analyze each file to detect new modules that are not yet deployed.
            for (File f : files) {
                // already deployed ?
                if (deployed.containsKey(f)) {
                    // yes, then check other files
                    continue;
                }

                // This module has failed previously ?
                if (failed.contains(f)) {
                    // If the module hasn't been updated, no need to deploy it again as it will fails again
                    if (!hasBeenUpdated(f)) {
                        continue;
                    }
                    // Cleanup the previous failure and try again the deployment
                    failed.remove(f);
                }


                // Else, get the deployable
                EZBArchive archive = ArchiveManager.getInstance().getArchive(f);
                if (archive == null) {
                    logger.warn("Ignoring invalid file ''{0}''", f);
                    continue;
                }
                EZBDeployable deployable;
                try {
                    deployable = DeployableHelper.getDeployable(archive);
                } catch (DeployableHelperException e) {
                    throw new EmbeddedException("Cannot get a deployable for the archive '" + archive + "'", e);
                }
                if (!bootInProgress) {
                    // wait that files are fully copied before deploying the files
                    try {
                        Thread.sleep(SLEEP_TIME);
                    } catch (InterruptedException e) {
                        throw new RuntimeException("Thread fail to sleep");
                    }
                }
                logger.debug("Detect a new Deployable ''{0}'' and deploying it.", deployable);

                // Now, deploy the file
                try {
                    deployer.deploy(deployable);
                } catch (EZBDeployerException e) {
                    // Deployment of this deployable has failed
                    failed.add(f);
                    throw new EmbeddedException("Cannot deploy the deployable '" + deployable + "'", e);
                } catch (RuntimeException e) {
                    // Runtime exception but deploiement has failed
                    failed.add(f);
                    throw new EmbeddedException("RuntimeException when deploying the deployable '" + deployable + "'", e);
                }

                // deployed is ok
                deployed.put(f, deployable);
            }
        }
    }



    /**
     * Check if the given file has been updated since the last check.
     * @param file the file to test
     * @return true if the archive has been updated
     */
    protected boolean hasBeenUpdated(final File file) {

        // get lastmodified for this URL
        long previousLastModified = 0;
        Long l = modifiedFiles.get(file);
        if (l != null) {
            previousLastModified = l.longValue();
        }

        long updatedModified = getLastModified(file);

        // first check. nothing to do
        if (previousLastModified == 0) {
            // Store initial time
            modifiedFiles.put(file, Long.valueOf(updatedModified));
            return false;
        }
        // URL has been updated since the last time
        if (updatedModified > previousLastModified) {
            modifiedFiles.put(file, Long.valueOf(updatedModified));
            return true;
        }

        return false;
    }

    /**
     * Check if the current deployables that are deployed have been updated.
     * If it is the case, undeploy them and then deploy it again (except for EJB3 Deployable where there is a stop/start).
     * @throws EmbeddedException if the redeployment fails
     */
    protected void checkModifiedDeployables() throws EmbeddedException {
        // Get list of files that are deployed
        Set<File> files = deployed.keySet();

        // Nothing to do if no modules are deployed.
        if (files == null) {
            return;
        }

        // For each deployed module that is not an EJB3, check if the module has been updated
        for (File f : files) {
            EZBDeployable deployable = deployed.get(f);

            // Not yet deployed ?
            if (deployable == null) {
                continue;
            }

            // EJB3 are managed in a different way
            if (deployable instanceof EJB3Deployable) {
                continue;
            }

            // File has been removed
            if (!f.exists()) {
                // undeploy
                logger.info("Deployable ''{0}'' has been removed on the filesystem, undeploy it", deployable);
                try {
                    deployer.undeploy(deployable);
                } catch (EZBDeployerException e) {
                    logger.error("Undeploy of the deployable '" + deployable + "' has failed", e);
                    failed.add(f);
                } finally {
                    // even in error case, the file should have been removed
                    deployed.remove(f);
                }
                continue;
            }

            // Update has been detected, need to undeploy and then to deploy again
            if (hasBeenUpdated(f)) {
                logger.info("Deployable ''{0}'' has been updated, reloading it", deployable);
                try {
                    deployer.undeploy(deployable);
                } catch (EZBDeployerException e) {
                    logger.error("Undeploy of the deployable '" + deployable + "' has failed", e);
                    // Deployment has failed, it is now undeployed
                    deployed.remove(f);
                    failed.add(f);
                }

                // Get a new deployable
                EZBArchive archive = ArchiveManager.getInstance().getArchive(f);
                if (archive == null) {
                    logger.warn("Ignoring invalid file ''{0}''", f);
                    continue;
                }
                EZBDeployable newDeployable;
                try {
                    newDeployable = DeployableHelper.getDeployable(archive);
                } catch (DeployableHelperException e) {
                    logger.error("Cannot get a deployable for the archive '" + archive + "'", e);
                    continue;
                }
                try {
                    deployer.deploy(newDeployable);
                } catch (EZBDeployerException e) {
                    // Deployment of this deployable has failed
                    failed.add(f);
                    throw new EmbeddedException("Cannot redeploy the deployable '" + deployable + "'.", e);
                }
            }

        }

    }


    /**
     * Check a container (and its archive) and see if there is a need to reload
     * the container.
     * @param container the container to monitor.
     */
    @Override
    protected void checkContainer(final EZBContainer container) {

        // get archive
        EZBArchive archive = container.getArchive();

        // Get URL
        URL url = null;
        try {
            url = archive.getURL();
        } catch (EZBArchiveException e1) {
            logger.warn("Cannot get URL on the container {0}", archive.getName());
            return;
        }
        File file = urlToFile(url);
        // No file archive, means that it has been removed
        if (!file.exists()) {
            logger.info("Archive ''{0}'' has been removed, then the associated EJB3 container is stopping", archive.getName());
            try {
                container.stop();
                getEmbedded().removeContainer(container);
            } finally {
                deployed.remove(file);
            }

            return;
        }

        // container was modified, need to relaunch it
        if (hasBeenUpdated(file)) {
            logger.info("Container with archive {0} was modified. Reloading...", archive.getName());
            try {
                container.stop();
                getEmbedded().removeContainer(container);
            } finally {
                deployed.remove(file);
            }
            try {
                container.start();
            } catch (EZBContainerException e) {
                logger.error("Error while restarting archive {0}.", archive.getName());
            }
        }

    }



    /**
     * Receives a stop order.
     */
    @Override
    public void stopOrder() {
        this.stopped = true;
    }
}
